/*
 * Copyright (C) 2002 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

/* Datasource object represent a source of PCM data. */

/* This is (and should be) mainly used by chunk.c */

#ifndef DATASOURCE_H_INCLUDED
#define DATASOURCE_H_INCLUDED

#include <gtk/gtk.h>
#include "gtkfiles.h"
#include "statusbar.h"

#define DATASOURCE(obj) GTK_CHECK_CAST(obj,datasource_get_type(),Datasource)
#define DATASOURCE_CLASS(klass) GTK_CHECK_CLASS_CAST(klass,datasource_get_type(),DatasourceClass)
#define IS_DATASOURCE(obj) GTK_CHECK_TYPE(obj,datasource_get_type())

#ifndef HAVE_LIBSNDFILE
#define SNDFILE void
#else
#include <sndfile.h>
#endif

/* struct describing the sample format. */

typedef struct {

     guint32 samplerate;         /* Samples/sec */
     guint samplesize;           /* bytes (1,2,3 or 4)*/
     gint channels;              /* Number of channels (1-16) */
     gboolean sign;              /* TRUE=signed, FALSE=unsigned */

     gint samplebytes;           /* should always be samplesize*channels */

} Dataformat;

/* Values for the type member of the Datasource struct. Determines how to 
 * actually get the data. */

#define DATASOURCE_REAL              0 /* Data stored directly in memory */
#define DATASOURCE_REAL_FLOAT        1 /* Normalized sample_t values stored in
					  memory */
#define DATASOURCE_VIRTUAL           2 /* The data is stored in a (read-only)
					* file. */ 
#define DATASOURCE_TEMPFILE          3 /* Data stored in a file that will be
			                * removed when the chunk is
			                * destroyed. */ 
#define DATASOURCE_TEMPFILE_FLOAT    4 /* Same as DATASOURCE_TEMPFILE
                                        * but the stored values are
					* sample_t values. */
#define DATASOURCE_SILENCE           5 /* Silence (duh) */
#define DATASOURCE_SNDFILE           6 /* The data is read using libSndfile */
#define DATASOURCE_SNDFILE_TEMPORARY 7 /* Same as DATASOURCE_SNDFILE but will be
                                        * removed when the chunk is
                                        * destroyed. */ 
#define DATASOURCE_REF               8 /* Reference to another Datasource.*/
#define DATASOURCE_CLONE             9 /* Clone of another Datasource. The 
					* Clone can have different PCM
                                        * format than the original. */ 
#define DATASOURCE_BYTESWAP          10/* Byteswapped clone. */

struct _Datasource;

struct _Datasource {

     GtkObject object;

     gint type;

     Dataformat format;

     off_t length;             /*Number of samples */
     off_t size;               /* Number of PCM bytes in file. */

     gboolean fake_pcm;          /* TRUE if the data is not real pcm
				    and reading as float would be more
				    efficient.. */
     
     guint opencount;            /* To keep track of nested open/close calls */

     gint tag;                   /* Extra field... */

     /* Type-specific data */
     union {

	  /* --- DATASOURCE_TEMPFILE and DATASOURCE_VIRTUAL --- */
	  struct {
	       char *filename; 
	       off_t offset;  /* Offset of data in file (bytes) */
	       EFILE *handle;
	       off_t pos;     /* Current position in file (bytes) */
	  } virtual;

	  /* --- DATASOURCE_REAL ---*/
	  char *real;  /* Pointer into the memory buffer with the sound data. */
	  

	  /* --- DATASOURCE_SNDFILE and DATASOURCE_SNDFILE_TEMPORARY --- */
	  struct {
	       char *filename;
	       gboolean raw_readable;
	       SNDFILE *handle;
	       off_t pos; /* Current position in file (samples) */
	  } sndfile;

	  /* DATASOURCE_CLONE, DATASOURCE_REF and DATASOURCE_BYTESWAP */
	  struct _Datasource *clone;
     } data;
};

typedef struct _Datasource Datasource;

typedef struct {
     GtkObjectClass klass;
} DatasourceClass;

#ifdef USE_DOUBLE_SAMPLES

typedef gdouble sample_t;
#define sf_readf_sample_t sf_readf_double
#define sf_writef_sample_t sf_writef_double
#define convert_pcm_float convert_pcm_float_d
#define convert_pcm_float_array convert_pcm_float_array_d
#define convert_float_pcm convert_float_pcm_d
#define convert_float_pcm_array convert_float_pcm_array_d

#else

typedef gfloat sample_t;
#define sf_readf_sample_t sf_readf_float
#define sf_writef_sample_t sf_writef_float
#define convert_pcm_float convert_pcm_float_f
#define convert_pcm_float_array convert_pcm_float_array_f
#define convert_float_pcm convert_float_pcm_f
#define convert_float_pcm_array convert_float_pcm_array_f

#endif

GtkType datasource_get_type(void);

/* ------------ 
 * CONSTRUCTORS 
 * ------------ */


/* Create data source from another Datasource object. The format of the data 
 * will be that of the format argument, regardless of what the other source
 * specifies. This makes it possible to override the format when you know it is
 * wrong. 
 */
Datasource *datasource_clone_df(Datasource *source, Dataformat *format);


/* Returns a byte-swapped clone of another Datasource. */
Datasource *datasource_byteswap(Datasource *source);


/* Create a data source from a memory buffer of data. The buffer will be 
 * g_free:d when the data source is destroyed and the data should not be 
 * modified by the caller any more after this call.
 */
Datasource *datasource_new_from_data(void *data, Dataformat *format, 
				     guint32 size);


/* Create a new datasource containing silence with a certain format and the 
 * specified number of samples  
 */
Datasource *datasource_new_silent(Dataformat *format, off_t samples);



/* ---------------------
 * DATA ACCESS FUNCTIONS
 * --------------------- */



/* Open the data source. Must be called before any of the other data access 
 * functions. May be called more than one time for the same Datasource.
 */

gboolean datasource_open(Datasource *source);



/* Close the data source. Must be called once for each call to 
 * datasource_open. 
 */

void datasource_close(Datasource *source);



/* Read one/many PCM/float sample from the Data source into the buffer. The
 * sample data will always be in host endian order.  
 */
gboolean datasource_read(Datasource *source, off_t sampleno, gpointer buffer);
gboolean datasource_read_fp(Datasource *source, off_t sampleno, 
			    sample_t *buffer);
/* Returns number of bytes read (0 on error). */
guint datasource_read_array(Datasource *source, off_t sampleno, 
			    guint size, gpointer buffer);
/* Returns number of (multi-channel) samples read (0 on error). */
guint datasource_read_array_fp(Datasource *source, off_t sampleno,
			       guint samples, sample_t *buffer);






/* -------------------
 * CONVERSION ROUTINES 
 * ------------------- */


/* These convert one or many samples between pcm formats and sample_t 
 * NOTE: These convert SINGLE-CHANNEL samples so you have to multiply the
 * count with the number of channels if you want to read full samples. 
 * NOTE 2: These routines assumes the data to be in host endian order.
 */

sample_t maximum_float_value(Dataformat *fmt);

/* These functions should be used in the program, they are #defined above:
sample_t convert_pcm_float(void *pcm, Dataformat *pcm_format);
void convert_pcm_float_array(void *pcm, Dataformat *pcm_format, 
			       sample_t *result, guint count);
void convert_float_pcm(sample_t f, void *result, Dataformat *result_format);
void convert_float_pcm_array(sample_t *f_buf, void *result, 
			     Dataformat *result_format, guint count);
*/

gfloat convert_pcm_float_f(void *pcm, Dataformat *pcm_format);
void convert_pcm_float_array_f(void *pcm, Dataformat *pcm_format, 
			       gfloat *result, guint count);
void convert_float_pcm_f(gfloat f, void *result, Dataformat *result_format);
void convert_float_pcm_array_f(gfloat *f_buf, void *result, 
			       Dataformat *result_format, guint count);

gdouble convert_pcm_float_d(void *pcm, Dataformat *pcm_format);
void convert_pcm_float_array_d(void *pcm, Dataformat *pcm_format, 
			       gdouble *result, guint count);
void convert_float_pcm_d(gdouble f, void *result, Dataformat *result_format);
void convert_float_pcm_array_d(gdouble *f_buf, void *result, 
			       Dataformat *result_format, guint count);


void datasource_conversion_selftest(void);


/* -------
 * VARIOUS 
 * ------- */


/* Returns the number of active Datasource objects */
guint datasource_count(void);



/* This dumps the PCM contents of the datasource into a file. You don't need
 * to open the datasource when using this function. The StatusBar must
 * already be in progress mode. 
 */

gboolean datasource_dump(Datasource *ds, off_t position, off_t length,
			 EFILE *file, gboolean bigendian, StatusBar *bar);


/* This reads all the data of the datasource into a memory buffer and converts
 * it into a memory datasource.
 */

gboolean datasource_realize(Datasource *ds);

/* This function will unlink a file. If any of the Datasources depend on this
 * file, the file will be moved or copied into a temporary file before the
 * original name is unlinked.
 */

gboolean datasource_backup_unlink(gchar *filename);


/* Returns TRUE if the two formats are equal */

gboolean dataformat_equal(Dataformat *f1, Dataformat *f2);

/* If the datasource can be tricked into another
 * samplesize/signedness without having to create a temporary file,
 * returns a new datasource with a datasorces data converted to other 
 * samplesize/signedness. Otherwise returns null. 
 */

Datasource *datasource_try_convert_samplesize(Datasource *ds, guint ssize, 
					      gboolean sign);


#endif
