/*
 * Copyright (C) 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <config.h>

#include <glib.h>
#include "sound.h"
#include "player.h"
#include "main.h"
#include "um.h"


/* Info about currently playing sound. */
static ChunkHandle *ch=NULL; /* NULL == stopped */
static off_t playstart, playend, curpos, realpos, timerel;
static guint32 warpcount; /* Not used for small loops */
static gboolean loop, small_loop;
static GTimeVal start_time;
static guint32 zeroes;

/* Output buffer */
static gchar player_buf[4096];
static gint player_bufsize,player_bufpos;


static off_t player_get_real_pos_main(void);

/* Check if the sound driver wants more data and send it.
 * Returns FALSE if it had nothing to do.
 * Since we want to be sure that we don't miss sound at the end, we send some
 * silence before stopping the sound. */
gboolean player_work(void)
{
     guint32 i;
     if (!ch || !output_want_data()) return FALSE;
     if (player_bufpos == player_bufsize) {
	  if (!loop && curpos==playend) {
	       /* Emit up to 1/2 second of silence before stopping output. */
	       if (zeroes < ch->format.samplerate*ch->format.samplebytes/2) {
		    memset(player_buf,0,sizeof(player_buf));
		    i = sizeof(player_buf);
		    i -= i%ch->format.samplebytes;
		    zeroes += output_play(player_buf,i);
	       }
	       if (output_play(NULL,0)==0 && player_get_real_pos()==playend) {
		    output_stop();
		    chunk_close(ch);
		    gtk_object_unref(GTK_OBJECT(ch));
		    ch=NULL;
	       } 
	       realpos = player_get_real_pos_main(); /* Underrun check */
	       return FALSE;
	  }
	  player_bufpos = 0;
	  if (!small_loop) {
	       i = MIN(sizeof(player_buf),(playend-curpos)*ch->format.samplebytes);
	       player_bufsize = chunk_read_array(ch,curpos,i,player_buf);
	       curpos += player_bufsize/ch->format.samplebytes;
	       if (curpos == playend) {
		    if (loop) { curpos = playstart; warpcount ++; }
		    else zeroes=0;
	       }
	  }
     }
     realpos = player_get_real_pos_main(); /* Underrun check */
     i = output_play(player_buf+player_bufpos,player_bufsize-player_bufpos);
     player_bufpos += i;
     return (i>0);
}


static void check_small_loop(void)
{
     off_t o;
     guint32 u,v;
     small_loop = FALSE;
     if (loop) {
	  o = (playend-playstart)*ch->format.samplebytes;
	  if (o <= sizeof(player_buf)) {
	       u = (guint32) o;
	       small_loop = TRUE;
	       chunk_read_array(ch,playstart,u,player_buf);
	       for (v=1; v < sizeof(player_buf)/u; v++) 
		    memcpy(player_buf+u*v,player_buf,u); 
	       /* printf("check_small_loop: u=%d, v=%d\n",u,v); */
	       player_bufsize = v*u;
	  }
     }
}


gboolean player_play(Chunk *chk, off_t spos, off_t epos, gboolean lp)
{
     player_stop();
     if (spos == epos) return TRUE;
     if (!output_supports_format(&(chk->format))) {
	  user_error("The sound format of this file is not supported for "
		     "playing.");
	  return TRUE;
     }
     output_select_format(&(chk->format));
     ch = chunk_open(chk);
     gtk_object_ref(GTK_OBJECT(chk));
     playstart = spos;
     playend = epos;
     loop = lp;
     player_bufsize = 0;
     player_bufpos = 0;
     check_small_loop();
     curpos = spos;
     timerel = curpos;
     warpcount = 0;
     g_get_current_time(&start_time);
     player_work();
     return FALSE;
}

off_t player_get_buffer_pos(void)
{
     return curpos;
}

/* This function calculates at what sample the listener is.
 * Since for most drivers the amount of buffering is unknown, I use clock time
 * to find out where we should be. After some time has passed, we should be
 * able to estimate the buffer amount by comparing the "clock" position and
 * the buffer position. Also, we check for overruns by checking for 
 * unreasonable values */
static off_t player_get_real_pos_main(void)
{
     GTimeVal tv,tv2;
     gfloat t;
     off_t rp,bp;
     guint32 rw,bw;
     
     /* Special case - we're called after player_work has closed the Chunk. */
     if (ch == NULL) return playend;


     g_get_current_time(&tv);
     timeval_subtract(&tv2,&tv,&start_time);
     t = tv2.tv_sec + 0.000001 * tv2.tv_usec;
     rp = timerel + (off_t)(t*((gfloat)(ch->format.samplerate)));
     if (rp>playend && !loop) rp=playend;     

     /* Calculate the number of times the "real" position has warped (rw) 
      * and compensate the position (rp) */
     if (rp>playend) {
	  rw = (rp-playend)/(playend-playstart) + 1;
	  rp -= (playend-playstart) * rw;
     } else {
	  rw = 0;
     }

     /* Special case - small loop */
     if (loop && small_loop)
	  return rp;    

     /* Special case - reached the end */
     if (curpos == playend && !loop) return rp;

     /* Calculate the first sample position that we are sure can't have been 
      * heard yet. */
     bp = curpos - (player_bufsize-player_bufpos)/ch->format.samplebytes -
	  output_passive_buffer()/ch->format.samplebytes - 
	  (off_t)(output_play_min_delay_time()*((float)ch->format.samplerate));
     bw = warpcount;
     if (bp < playstart) {
	  if (bw == 0) bp = playstart;
	  else {
	       /* FIXME: Replace this loop */
	       while (bp < playstart) {
		    bp += (playend - playstart);
		    bw --;
	       }
	  }
     }
     g_assert(bw >= 0 && bp >= playstart && bp < playend);

     /* Check if the timer-baser position is ahead of us. 
      * Then we must've had an underrun. */
     if (bw<rw || (bw==rw && bp<rp)) {
	  /* Assuming that the playback has just restarted after an underrun.*/
	  memcpy(&start_time,&tv,sizeof(start_time));
	  rp = timerel = bp;
	  warpcount = 0;
     } 

     return rp;
}

off_t player_get_real_pos(void)
{
     return realpos;
}

void player_get_range(off_t *startpos, off_t *endpos)
{
     *startpos = playstart;
     *endpos = playend;
}

void player_set_buffer_pos(off_t pos)
{
     if (pos != curpos) {
	  curpos = pos;
	  if (curpos > playend) playend = ch->length;
	  if (curpos < playstart) playstart = curpos;
	  g_get_current_time(&start_time);
	  timerel = pos;
	  warpcount = 0;
	  output_clear_buffers();
	  player_bufsize = player_bufpos = 0;
	  player_work();
     }
}

void player_replace(Chunk *chunk, off_t start, off_t end, off_t pos)
{
     ChunkHandle *c;
     off_t rp = player_get_real_pos();
     off_t oldpos = curpos;

     /* Special case - we're playing a loop and the sound has looped, but 
      * because of the output delay the cursor and the audible sound hasn't 
      * come there yet. In that case, we're forced to flush buffers etc 
      * to avoid sending out the old loop to the listener. */
     if (loop && rp > oldpos) {
	  player_play(chunk,rp,end,TRUE);
	  playstart = start;
	  return;
     }

     /* Special case - the resulting loop will be very small */    
     if (loop && 
	 (end - start)*chunk->format.samplebytes <= sizeof(player_buf)) {
	  player_play(chunk,start,end,TRUE);
	  g_assert(small_loop);
     }

     /* Special case - the current loop is very small */
     if (loop && small_loop) {
	  player_play(chunk,rp,end,TRUE);
	  playstart = start;
	  return;
     }


     c = chunk_open(chunk);
     if (c == NULL) return;
     chunk_close(ch);
     gtk_object_unref(GTK_OBJECT(ch));
     ch = c;
     gtk_object_ref(GTK_OBJECT(ch));

     playstart = start;
     playend = end;
     curpos = pos;

     /* Update timerel */
     timerel = rp + pos - oldpos;
}


gboolean player_playing(void)
{
     return (ch != NULL);
}



void player_stop(void)
{
     if (ch == NULL) return;
     chunk_close(ch);
     gtk_object_unref(GTK_OBJECT(ch));
     ch = NULL;
     output_clear_buffers();
     output_stop();
}

gboolean player_looping(void)
{
     return loop;
}

void player_nudge(gfloat seconds)
{
     off_t samps,newpos;
     samps = seconds * ch->format.samplerate;
     newpos = player_get_real_pos() + samps;
     if (newpos < 0) newpos = 0;
     player_set_buffer_pos(newpos);
}
